use std::marker::PhantomData;

use super::*;

pub struct ControlFlowClass<C>(ControlFlow<(), C>);

impl<C> WeakFunctor for ControlFlowClass<C> {
    type F<'a, A: 'a> =  ControlFlow<A, C>
    where
        Self: 'a;
}

impl<C> Functor for ControlFlowClass<C> {
    fn fmap<'a, A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::F<'a, A>) -> Self::F<'a, B>
    where
        Self: 'a,
    {
        match fa {
            ControlFlow::Continue(c) => ControlFlow::Continue(c),
            ControlFlow::Break(a) => ControlFlow::Break(f(a)),
        }
    }
}

impl<C> Pure for ControlFlowClass<C> {
    fn pure<'a, A: 'a>(a: A) -> Self::F<'a, A>
    where
        Self: 'a,
    {
        ControlFlow::Break(a)
    }
}

/// Next [AIterative] state, wrapped.
pub type AIterativeWrapped<'a, I> = <<I as AIterative<'a>>::T as WeakFunctor>::F<
    'a,
    ControlFlow<<I as AIterative<'a>>::B, (<I as AIterative<'a>>::A, I)>,
>;

/// Value passed to [`Monad::iterate_argument`].
/// `A` prefix meaning "with an **a**rgument" given current implementation
/// relying on the argument getting passed around.
pub trait AIterative<'a>: 'a + Sized {
    /// [`ControlFlow::Continue`].
    type A: 'a;
    /// [`ControlFlow::Break`].
    type B: 'a;
    /// Corresponding [`WeakFunctor`].
    type T: 'a + ?Sized + WeakFunctor;
    /// Get next state.
    fn next(self, a: Self::A) -> AIterativeWrapped<'a, Self>;
}

pub struct BindableMut<T: ?Sized, A, B, F>(F, PhantomData<A>, PhantomData<B>, PhantomData<T>);

impl<
        'a,
        T: 'a + ?Sized + Functor,
        A: 'a,
        B: 'a,
        F: 'a + FnMut(A) -> T::F<'a, ControlFlow<B, A>>,
    > BindableMut<T, A, B, F>
{
    pub fn new(f: F) -> Self {
        BindableMut(f, PhantomData, PhantomData, PhantomData)
    }
}

impl<
        'a,
        T: 'a + ?Sized + Functor,
        A: 'a,
        B: 'a,
        F: 'a + FnMut(A) -> T::F<'a, ControlFlow<B, A>>,
    > AIterative<'a> for BindableMut<T, A, B, F>
{
    type A = A;
    type B = B;
    type T = T;

    fn next(mut self, a: Self::A) -> AIterativeWrapped<'a, Self> {
        let fa = self.0(a);
        T::fmap(
            move |state| match state {
                ControlFlow::Continue(next_a) => ControlFlow::Continue((next_a, self)),
                ControlFlow::Break(b) => ControlFlow::Break(b),
            },
            fa,
        )
    }
}

/// Next [Iterative] state, wrapped.
pub type IterativeWrapped<'a, I> =
    <<I as Iterative<'a>>::T as WeakFunctor>::F<'a, ControlFlow<<I as Iterative<'a>>::B, I>>;

/// Value passed to [`Monad::iterate`].
pub trait Iterative<'a>: 'a + Sized {
    /// [`ControlFlow::Break`].
    type B: 'a;
    /// Corresponding [`WeakFunctor`].
    type T: 'a + ?Sized + WeakFunctor;
    /// Get next state.
    fn next(self) -> IterativeWrapped<'a, Self>;
}

pub struct ArgumentedIterative<A, F>(pub A, pub F);

impl<'a, A: 'a, F: AIterative<'a, A = A>> Iterative<'a> for ArgumentedIterative<A, F>
where
    F::T: Functor,
{
    type B = F::B;

    type T = F::T;

    fn next(self) -> IterativeWrapped<'a, Self> {
        Self::T::fmap(
            |state| match state {
                ControlFlow::Continue((next_a, next_f)) => {
                    ControlFlow::Continue(ArgumentedIterative(next_a, next_f))
                }
                ControlFlow::Break(b) => ControlFlow::Break(b),
            },
            self.1.next(self.0),
        )
    }
}
